@@ -29,6 +29,7 @@
 #undef GDK_MULTIHEAD_SAFE
 #endif
 
+#include "netk-private.h"
 #include "netk-pager.h"
 #include "netk-workspace.h"
 #include "netk-window.h"
@@ -37,6 +38,8 @@
 
 #define N_SCREEN_CONNECTIONS 11
 
+#define NETK_PAGER_GET_PRIVATE(o)    (G_TYPE_INSTANCE_GET_PRIVATE ((o), NETK_TYPE_PAGER, NetkPagerPrivate))
+
 struct _NetkPagerPrivate
 {
     NetkScreen *screen;
@@ -44,6 +47,7 @@ struct _NetkPagerPrivate
     int n_rows;                 /* really columns for vertical orientation */
     NetkPagerDisplayMode display_mode;
     gboolean show_all_workspaces;
+    gboolean workspace_scrolling;
     GtkShadowType shadow_type;
 
     GtkOrientation orientation;
@@ -64,6 +68,8 @@ struct _NetkPagerPrivate
     GdkPixbuf *bg_cache;
 
     int layout_manager_token;
+    int ws_activate;
+    NetkWorkspace *ws_to_activate;
 };
 
 #define POINT_IN_RECT(xcoord, ycoord, rect) \
@@ -110,46 +116,36 @@ static void netk_pager_clear_drag (NetkPager * pager);
 static GdkPixbuf *netk_pager_get_background (NetkPager * pager, int width,
                                              int height);
 
-static gpointer parent_class;
+static GObjectClass *parent_class;
 
 GType
 netk_pager_get_type (void)
 {
-    static GType object_type = 0;
+    static GType type = G_TYPE_INVALID;
 
-    g_type_init ();
-
-    if (!object_type)
+    if (G_UNLIKELY (type == G_TYPE_INVALID))
     {
-        static const GTypeInfo object_info = {
-            sizeof (NetkPagerClass),
-            (GBaseInitFunc) NULL,
-            (GBaseFinalizeFunc) NULL,
-            (GClassInitFunc) netk_pager_class_init,
-            NULL,               /* class_finalize */
-            NULL,               /* class_data */
-            sizeof (NetkPager),
-            0,                  /* n_preallocs */
-            (GInstanceInitFunc) netk_pager_init,
-            NULL		/* value_table */
-        };
-
-        object_type =
-            g_type_register_static (GTK_TYPE_WIDGET, "NetkPager",
-                                    &object_info, 0);
+        type = _netk_g_type_register_simple (GTK_TYPE_WIDGET,
+                                             "NetkPager",
+                                             sizeof (NetkPagerClass),
+                                             netk_pager_class_init,
+                                             sizeof (NetkPager),
+                                             netk_pager_init,
+                                             0);                                    
     }
 
-    return object_type;
+    return type;
 }
 
 static void
 netk_pager_init (NetkPager * pager)
 {
-    pager->priv = g_new0 (NetkPagerPrivate, 1);
-
+    pager->priv = NETK_PAGER_GET_PRIVATE (pager);
+    
     pager->priv->n_rows = 1;
     pager->priv->display_mode = NETK_PAGER_DISPLAY_CONTENT;
     pager->priv->show_all_workspaces = TRUE;
+    pager->priv->workspace_scrolling = TRUE;
     pager->priv->shadow_type = GTK_SHADOW_NONE;
     pager->priv->orientation = GTK_ORIENTATION_HORIZONTAL;
     pager->priv->workspace_size = 48;
@@ -164,6 +160,8 @@ netk_pager_class_init (NetkPagerClass * klass)
 {
     GObjectClass *object_class = G_OBJECT_CLASS (klass);
     GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
+    
+    g_type_class_add_private (klass, sizeof (NetkPagerPrivate));
 
     parent_class = g_type_class_peek_parent (klass);
 
@@ -196,8 +194,6 @@ netk_pager_finalize (GObject * object)
         pager->priv->bg_cache = NULL;
     }
 
-    g_free (pager->priv);
-
     G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
@@ -266,6 +262,7 @@ netk_pager_size_request (GtkWidget * widget, GtkRequisition * requisition)
     int size;
     int n_rows;
     int focus_width;
+    NetkWorkspace *space;
 
     pager = NETK_PAGER (widget);
 
@@ -274,12 +271,22 @@ netk_pager_size_request (GtkWidget * widget, GtkRequisition * requisition)
     g_assert (pager->priv->n_rows > 0);
     spaces_per_row =
         (n_spaces + pager->priv->n_rows - 1) / pager->priv->n_rows;
+    space = netk_screen_get_workspace (pager->priv->screen, 0);
 
     if (pager->priv->orientation == GTK_ORIENTATION_VERTICAL)
     {
-        screen_aspect =
-            (double) netk_screen_get_height (pager->priv->screen) /
-            (double) netk_screen_get_width (pager->priv->screen);
+        if (space)
+        {
+            screen_aspect =
+                  (double) netk_workspace_get_height (space) /
+                  (double) netk_workspace_get_width (space);
+        }
+        else
+        {
+            screen_aspect =
+                  (double) netk_screen_get_height (pager->priv->screen) /
+                  (double) netk_screen_get_width (pager->priv->screen);
+        }
 
         /* TODO: Handle NETK_PAGER_DISPLAY_NAME for this case */
 
@@ -303,9 +310,18 @@ netk_pager_size_request (GtkWidget * widget, GtkRequisition * requisition)
     }
     else
     {
-        screen_aspect =
-            (double) netk_screen_get_width (pager->priv->screen) /
-            (double) netk_screen_get_height (pager->priv->screen);
+        if (space) 
+        {
+            screen_aspect =
+                  (double) netk_workspace_get_width (space) /
+                  (double) netk_workspace_get_height (space);
+        } 
+        else 
+        {
+            screen_aspect =
+                  (double) netk_screen_get_width (pager->priv->screen) /
+                  (double) netk_screen_get_height (pager->priv->screen);
+        }
 
         if (pager->priv->show_all_workspaces)
         {
@@ -344,7 +360,7 @@ netk_pager_size_request (GtkWidget * widget, GtkRequisition * requisition)
                 other_dimension_size = MAX (other_dimension_size, w);
             }
 
-            g_object_unref (layout);
+            g_object_unref (G_OBJECT (layout));
 
             other_dimension_size += 2;
         }
@@ -394,16 +410,14 @@ netk_pager_size_allocate (GtkWidget * widget, GtkAllocation * allocation)
     if (pager->priv->orientation == GTK_ORIENTATION_VERTICAL)
     {
         if (pager->priv->show_all_workspaces)
-            workspace_size =
-                (width - (pager->priv->n_rows - 1)) / pager->priv->n_rows;
+            workspace_size = (width - (pager->priv->n_rows - 1)) / pager->priv->n_rows;
         else
             workspace_size = width;
     }
     else
     {
         if (pager->priv->show_all_workspaces)
-            workspace_size =
-                (height - (pager->priv->n_rows - 1)) / pager->priv->n_rows;
+            workspace_size = (height - (pager->priv->n_rows - 1)) / pager->priv->n_rows;
         else
             workspace_size = height;
     }
@@ -926,7 +940,7 @@ netk_pager_draw_workspace (NetkPager * pager,
                          rect->x + (rect->width - w) / 2,
                          rect->y + (rect->height - h) / 2, layout);
 
-        g_object_unref (layout);
+        g_object_unref (G_OBJECT (layout));
     }
 }
 
@@ -1229,6 +1243,10 @@ netk_pager_scroll_event (GtkWidget * widget, GdkEventScroll * event)
     g_return_val_if_fail (widget != NULL, FALSE);
 
     pager = NETK_PAGER (widget);
+    
+    if (!pager->priv->workspace_scrolling)
+	return FALSE;
+    
     n = netk_screen_get_workspace_count (pager->priv->screen);
     active =
         netk_workspace_get_number (netk_screen_get_active_workspace
@@ -1279,6 +1297,67 @@ netk_pager_focus (GtkWidget * widget, GtkDirectionType direction)
     return GTK_WIDGET_CLASS (parent_class)->focus (widget, direction);
 }
 
+/* dragging something over the pager */
+static gboolean
+netk_pager_drag_timeout (NetkPager *pager)
+{
+    netk_workspace_activate (pager->priv->ws_to_activate);
+    pager->priv->ws_activate = 0;
+    pager->priv->ws_to_activate = NULL;
+    return FALSE;
+}
+
+static void
+netk_pager_drag_leave (GtkWidget * widget, GdkDragContext * context,
+                       guint time, gpointer data)
+{
+    NetkPager *pager = NETK_PAGER (widget);
+
+    if (pager->priv->ws_activate != 0)
+    {
+        g_source_remove (pager->priv->ws_activate);
+        pager->priv->ws_activate = 0;
+        pager->priv->ws_to_activate = NULL;
+    }
+}
+
+static gboolean
+netk_pager_drag_motion (GtkWidget * widget, GdkDragContext * context, gint x,
+                        gint y, guint time, gpointer data)
+{
+    NetkPager *pager = NETK_PAGER (widget);
+    NetkWorkspace *current_ws, *new_ws;
+    int num;
+
+    current_ws = netk_screen_get_active_workspace (pager->priv->screen);
+
+    num = workspace_at_point (pager, x, y, NULL, NULL);
+    
+    if (num != -1) 
+    {
+        new_ws = netk_screen_get_workspace (pager->priv->screen, num);
+        if (new_ws != current_ws && new_ws != pager->priv->ws_to_activate)
+        {
+            if (pager->priv->ws_activate != 0)
+            {
+                g_source_remove (pager->priv->ws_activate);
+            }
+            pager->priv->ws_to_activate = new_ws;
+            pager->priv->ws_activate = 
+                g_timeout_add (1000, (GSourceFunc)netk_pager_drag_timeout, pager);
+        }
+    }
+    else
+    {
+        g_source_remove (pager->priv->ws_activate);
+        pager->priv->ws_activate = 0;
+        pager->priv->ws_to_activate = NULL;
+    }
+    gdk_drag_status (context, 0, time);
+
+    return TRUE;
+}
+
 void
 netk_pager_set_screen (NetkPager * pager, NetkScreen * screen)
 {
@@ -1298,6 +1377,12 @@ netk_pager_new (NetkScreen * screen)
 
     pager = g_object_new (NETK_TYPE_PAGER, NULL);
 
+    gtk_drag_dest_set (GTK_WIDGET (pager), 0, NULL, 0, 0);
+    g_signal_connect (pager, "drag_motion",
+                      G_CALLBACK (netk_pager_drag_motion), NULL);
+    g_signal_connect (pager, "drag_leave",
+                      G_CALLBACK (netk_pager_drag_leave), NULL);
+
     netk_pager_connect_screen (pager, screen);
 
     return GTK_WIDGET (pager);
@@ -1389,6 +1474,18 @@ netk_pager_set_show_all (NetkPager * pager, gboolean show_all_workspaces)
     gtk_widget_queue_resize (GTK_WIDGET (pager));
 }
 
+void
+netk_pager_set_workspace_scrolling (NetkPager * pager, gboolean workspace_scrolling)
+{
+    g_return_if_fail (NETK_IS_PAGER (pager));
+
+    if (pager->priv->workspace_scrolling == workspace_scrolling)
+        return;
+
+    pager->priv->workspace_scrolling = workspace_scrolling;
+    gtk_widget_queue_resize (GTK_WIDGET (pager));
+}
+
 void
 netk_pager_set_shadow_type (NetkPager * pager, GtkShadowType shadow_type)
 {
